/*
 * Copyright (C) 2013-2018 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 * Author: Tomasz Wlostowski <tomasz.wlostowski@cern.ch>
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */

#include <errno.h>
#include <string.h>

#include "mockturtle-rt.h"
#include <mockturtle-framework.h>
#include "wrtd-common.h"
#include "wrtd-fw-common.h"
#include "hw/fd_channel_regs.h"
#include "hw/fd_main_regs.h"

#include "fine-delay-init.h"

#include "loop-queue.h"

#define WRTD_OUT_HMQ 0
#define WRTD_OUT_RMQ 0

#define OUT_TIMEOUT 10

#ifdef SIMULATION
static const int is_simulation = 1;
#else
static const int is_simulation = 0;
#endif

struct wrtd_out_trigger triggers[FD_HASH_ENTRIES]; /**< list of triggers */
struct wrtd_out_trigger *ht[FD_HASH_ENTRIES]; /* hash table */
static unsigned int tlist_count = 0; /**< number of valid trigger entry
					in tlist */
static struct wrtd_out_channel wrtd_out_channels[FD_NUM_CHANNELS]; /**< Output
								      state
								      array */
static struct wrtd_out wrtd_out_device;

enum wrtd_in_wr_link {
	WR_LINK_OFFLINE = 1,
	WR_LINK_ONLINE,
	WR_LINK_SYNCING,
	WR_LINK_SYNCED,
};

static int wr_state;

static int wr_present(void)
{
	return !!(dp_readl(FD_REG_TCR) & FD_TCR_WR_PRESENT);
}

static int wr_link_up()
{
	return dp_readl ( FD_REG_TCR ) & FD_TCR_WR_LINK;
}

static int wr_time_locked()
{
	return dp_readl ( FD_REG_TCR ) & FD_TCR_WR_LOCKED;
}

static int wr_time_ready()
{
	return 1;
}

static void wr_enable_lock(int enable)
{
	if(enable)
		dp_writel(FD_TCR_WR_ENABLE, FD_REG_TCR);
	else
		dp_writel(0, FD_REG_TCR);
}

/**
 * It updates the White-Rabbit link status
 */
static void wr_update_link(void)
{
	int i;

	if (is_simulation) {
		wr_state = WR_LINK_SYNCED;
		return;
	}

	switch(wr_state) {
	case WR_LINK_OFFLINE:
		for(i = 0; i < FD_NUM_CHANNELS; i++)
			wrtd_out_channels[i].config.flags |= WRTD_NO_WR;
		if (wr_link_up())
			wr_state = WR_LINK_ONLINE;
		break;
	case WR_LINK_ONLINE:
		if (wr_time_ready()) {
			wr_state = WR_LINK_SYNCING;
			wr_enable_lock(1);
		}
		break;
	case WR_LINK_SYNCING:
		if (wr_time_locked()) {
			pr_debug("WR link up!\n\r");
			wr_state = WR_LINK_SYNCED;
		}
		break;
	case WR_LINK_SYNCED:
		for(i = 0; i < FD_NUM_CHANNELS; i++)
			wrtd_out_channels[i].config.flags &= ~WRTD_NO_WR;
		break;
	}

	if( wr_state != WR_LINK_OFFLINE && !wr_link_up() ) {
		pr_error("rt-out: WR sync lost\n\r");
		wr_state = WR_LINK_OFFLINE;
		wr_enable_lock(0);
	}
}

static int wr_is_timing_ok()
{
	return (wr_state == WR_LINK_SYNCED);
}


/**
 * Writes to FD output registers for output (out)
 */
static inline void fd_ch_writel(struct wrtd_out_channel *out, uint32_t value,
				 uint32_t reg)
{
	dp_writel(value , reg + out->base_addr);
}

/**
 * Reads from FD output registers for output (out)
 */
static inline uint32_t fd_ch_readl (struct wrtd_out_channel *out,  uint32_t reg)
{
	return dp_readl(reg + out->base_addr);
}


/**
 * Compare two trigger identifiers.
 * It does not use memcmp() because the shared memory use word addressing, so
 * it will not work if one of the variable is located on the shared memory
 */
static inline int trig_eq(struct wrtd_trig_id *id1, struct wrtd_trig_id *id2)
{
	return id1->system == id2->system &&
		id1->source_port == id2->source_port &&
		id1->trigger == id2->trigger;
}


/**
 * It looks for a trigger entry within the hash table
 */
static int wrtd_out_hash_table_find(struct wrtd_trig_id *tid)
{
	int i, hidx = wrtd_hash_func(tid);

	for (i = 0; i < FD_HASH_ENTRIES;
	     ++i, hidx = (hidx + 1) & (FD_HASH_ENTRIES - 1)) {
		if (trig_eq(tid, &ht[hidx]->id))
			return hidx;
	}

	return -1;
}


/**
 * It looks for a trigger entry within the hash table
 */
static inline struct wrtd_out_trigger *rtfd_trigger_entry_find(struct wrtd_trig_id *tid)
{
	int hidx = wrtd_out_hash_table_find(tid);

	return hidx >= 0 ? ht[hidx] : NULL;
}


/**
 * Adjusts the timestamp in-place by adding cycles/frac value
 */
static void ts_adjust_delay(struct wr_timestamp *ts, uint32_t cycles, uint32_t frac)
{
	ts->frac += frac;

	if (ts->frac & 0x1000)
		ts->ticks++;

	ts->frac &= 0xfff;
	ts->ticks += cycles;

	if (ts->ticks >= 125000000) {
		ts->ticks -= 125000000;
		ts->seconds++;
	}
}

/**
 * Puts a trigger message in the log buffer
 */
static void log_trigger(int type, int miss_reason,
			struct wrtd_out_channel *out,
			struct wrtd_trigger_entry *ent)
{
	struct wrtd_log_entry *log;
	struct trtl_fw_msg msg;
	int ret;

	if (!out && (type != WRTD_LOG_PROMISC))
		return;
	if (out && !(out->config.log_level & type))
		return;

	ret = mq_claim(TRTL_HMQ, WRTD_OUT_HMQ);
	if (ret == -EBUSY)
		return;

	mq_map_out_message(TRTL_HMQ, WRTD_OUT_HMQ, &msg);

	msg.header->msg_id = WRTD_OUT_ACTION_LOG;
	msg.header->len = sizeof(struct wrtd_log_entry) / 4;

	log = (struct wrtd_log_entry *)msg.payload;
	log->type = type;
	log->channel = out ? out->n : -1;
	log->miss_reason = miss_reason;
	log->seq = ent->seq;
	log->id = ent->id;
	log->ts = ent->ts;

	pr_debug("%s: %d %d\n\r", __func__, type, miss_reason);

	mq_send(TRTL_HMQ, WRTD_OUT_HMQ);
}


/**
 * Initializes an empty pulse queue
 */
static void pulse_queue_init(struct lrt_pulse_queue *p)
{
	p->head = 0;
	p->tail = 0;
	p->count = 0;
}


/**
 * Requests a new entry in a pulse queue. Returns pointer to the ne
 * entry or NULL if the queue is full.
 */
struct pulse_queue_entry *pulse_queue_push(struct lrt_pulse_queue *p)
{
	struct pulse_queue_entry *ent;

	if (p->count == FD_MAX_QUEUE_PULSES)
		return NULL;

	ent = &p->data[p->head];
	p->count++;
	p->head++;

	if (p->head == FD_MAX_QUEUE_PULSES)
		p->head = 0;

	return ent;
}


/**
 * Returns non-0 if pulse queue p contains any pulses.
 */
static inline int pulse_queue_empty(struct lrt_pulse_queue *p)
{
	return (p->count == 0);
}


/**
 * Returns the oldest entry in the pulse queue (or NULL if empty).
 */
static struct pulse_queue_entry* pulse_queue_front(struct lrt_pulse_queue *p)
{
	if (!p->count)
		return NULL;
	return &p->data[p->tail];
}


/**
 * Releases the oldest entry from the pulse queue.
 */
static void pulse_queue_pop(struct lrt_pulse_queue *p)
{
	p->tail++;

	if(p->tail == FD_MAX_QUEUE_PULSES)
		p->tail = 0;
	p->count--;
}

/**
 * Checks if the timestamp of the pulse (ts) does not violate the
 * dead time on the output out by comparing it with the last processed
 * pulse timestamp.
 */
static int check_dead_time(struct wrtd_out_channel *out,
			   struct wr_timestamp *ts)
{
	int delta_s = ts->seconds - out->priv.prev_pulse.seconds;
	int delta_c = ts->ticks - out->priv.prev_pulse.ticks;

	if (delta_c < 0) {
		delta_c += 125 * 1000 * 1000;
		delta_s--;
	}

	return (delta_s < 0 || delta_c < out->config.dead_time) ? 0 : 1;
}


/**
 * Checks if the timestamp of the last programmed pulse is lost because
 * of timeout
 */
static int check_output_timeout (struct wrtd_out_channel *out)
{
	struct lrt_pulse_queue *q = &out->priv.queue;
	struct pulse_queue_entry *pq_ent = pulse_queue_front(q);
	struct wr_timestamp tc;
	int delta;

	/*
	 * Read the current WR time, order is important: first seconds,
	 * then cycles (cycles get latched on reading secs register.
	 */
	tc.seconds = lr_readl(MT_CPU_LR_REG_TAI_SEC);
	tc.ticks = lr_readl(MT_CPU_LR_REG_TAI_CYCLES);

	if( out->stats.last_programmed.seconds > tc.seconds + OUT_TIMEOUT ) {
		pr_error("Enqueued timestamp very far in the future [id %"PRIx32":%"PRIx32":%"PRIx32"]. Dropping.",
			 pq_ent->trig.id.system,
			 pq_ent->trig.id.source_port,
			 pq_ent->trig.id.trigger);
		pr_error("Offending TS: %"PRIu64":%"PRIu32,
			  out->stats.last_programmed.seconds,
			  out->stats.last_programmed.ticks);
	}

	delta = tc.seconds - out->stats.last_programmed.seconds;
	delta *= 125 * 1000 * 1000;
	delta += tc.ticks - out->stats.last_programmed.ticks;

	/* Current time exceeds FD setpoint? */
	return (delta > 0);
}


/**
 * It updates the latency stats
 */
static void update_latency_stats(struct pulse_queue_entry *pq_ent)
{
	/* Read the time and calculate the latency */
	struct lrt_output_rule *rule = pq_ent->rule;
	__attribute__((unused))
		volatile uint32_t dummy = lr_readl (MT_CPU_LR_REG_TAI_SEC);
	int latency;

	if(!rule)
		return;
	latency = lr_readl (MT_CPU_LR_REG_TAI_CYCLES) - pq_ent->origin_cycles;
	if (latency < 0)
		latency += 125 * 1000 * 1000;

	if (latency > rule->latency_worst)
		rule->latency_worst = latency;

	if (rule->latency_avg_sum > 2000 * 1000 * 1000) {
		rule->latency_avg_sum = 0;
		rule->latency_avg_nsamples = 0;
	}

	rule->latency_avg_sum += latency;
	rule->latency_avg_nsamples++;
}


/**
 * Drop the given enqueued trigger
 */
static void drop_trigger(struct wrtd_out_channel *out,
			 struct pulse_queue_entry *pq_ent,
			 struct lrt_pulse_queue *q, int reason)
{
	out->priv.idle = 1;

	if (pulse_queue_empty(q))
	    return;

	/* Drop the pulse */
	pulse_queue_pop(q);
	pq_ent->rule->misses ++;

	switch (reason) {
	case WRTD_MISS_TIMEOUT:
		out->stats.miss_timeout++;
		break;
	case WRTD_MISS_NO_WR:
		out->stats.miss_no_timing ++;
		break;
	}

	out->stats.last_lost = pq_ent->trig;

	/* Disarm the FD output */
	fd_ch_writel(out, FD_DCR_MODE, FD_REG_DCR);

	if(out->config.state == OUT_ST_TEST_PENDING)
		out->config.state = OUT_ST_IDLE;

	log_trigger(WRTD_LOG_MISSED, reason, out, &pq_ent->trig);
}


/**
 * Output driving function. Reads pulses from the output queue,
 * programs the output and updates the output statistics.
 */
static void do_output (struct wrtd_out_channel *out)
{
	struct lrt_pulse_queue *q = &out->priv.queue;
	struct pulse_queue_entry *pq_ent = pulse_queue_front(q);
	uint32_t dcr = fd_ch_readl(out, FD_REG_DCR);
	struct wr_timestamp *ts;

	/* Check if the output has triggered */
	if(!out->priv.idle) {
		if (!wr_is_timing_ok()) {
			drop_trigger(out, pq_ent, q, WRTD_MISS_NO_WR);
		}
		else if (!(dcr & FD_DCR_PG_TRIG)) {
			/* Nope, armed but still waiting for trigger */
			if (check_output_timeout (out))
				drop_trigger(out, pq_ent, q, WRTD_MISS_TIMEOUT);
		} else {
			out->stats.last_executed = pq_ent->trig;
			pq_ent->rule->hits ++;
			pulse_queue_pop(q);
			out->stats.hits++;
			out->priv.idle = 1;
			out->config.flags |= WRTD_TRIGGERED;

			if(out->config.state == OUT_ST_TEST_PENDING)
				out->config.state = OUT_ST_IDLE;

			log_trigger (WRTD_LOG_EXECUTED, 0, out, &pq_ent->trig);
		}
		return;
	}

	/* Output is idle: check if there's something in the queue to execute */
	if (pulse_queue_empty(q))
		return;

	pq_ent = pulse_queue_front(q);
	ts = &pq_ent->trig.ts;

	if (!wr_is_timing_ok())
		drop_trigger(out, pq_ent, q, WRTD_MISS_NO_WR);


	/* Program the output start time */
	fd_ch_writel(out, ts->seconds, FD_REG_U_STARTL);
	fd_ch_writel(out, ts->ticks, FD_REG_C_START);
	fd_ch_writel(out, ts->frac, FD_REG_F_START);

	/* Adjust pulse width and program the output end time */
	ts->ticks += out->config.width_cycles;
	if (ts->ticks >= 125000000) {
		ts->ticks -= 125000000;
		ts->seconds++;
	}

	fd_ch_writel(out, ts->seconds, FD_REG_U_ENDL);
	fd_ch_writel(out, ts->ticks, FD_REG_C_END);
	fd_ch_writel(out, ts->frac, FD_REG_F_END);
	fd_ch_writel(out, 0, FD_REG_RCR);
	fd_ch_writel(out, FD_DCR_MODE, FD_REG_DCR);
	fd_ch_writel(out, FD_DCR_MODE | FD_DCR_UPDATE, FD_REG_DCR);
	fd_ch_writel(out, FD_DCR_MODE | FD_DCR_PG_ARM | FD_DCR_ENABLE,
		     FD_REG_DCR);

	ts->ticks += 1000;
	if (ts->ticks >= 125000000) {
		ts->ticks -= 125000000;
		ts->seconds++;
	}

	/*
	 * Store the last programmed timestamp (+ some margin) and mark
	 * the output as busy
	 */
	out->stats.last_programmed = *ts;
	out->priv.idle = 0;

	update_latency_stats (pq_ent);
}


/**
 * If the given trigger can be generated, then push it into the queue
 */
static void enqueue_trigger(int output, struct lrt_output_rule *rule,
			    struct wrtd_trig_id *id,
			    struct wr_timestamp *ts, int seq)
{
	struct wrtd_out_channel *out = &wrtd_out_channels[output];
	struct wr_timestamp adjusted = *ts;
	struct pulse_queue_entry *pq_ent;
	struct wrtd_trigger_entry ent;

	pr_debug("%s:%d ch %d 0x%x\n\r", __func__, __LINE__, output, rule->state);

	if(rule->state & HASH_ENT_DISABLED)
		return;

	if (!(out->config.flags & WRTD_ENABLED))
		return;

	ts_adjust_delay(&adjusted, rule->delay_cycles, rule->delay_frac);

	ent.ts = *ts;
	ent.id = *id;
	ent.seq = seq;

	if (!check_dead_time(out, &adjusted)) {
		rule->misses ++;
		out->stats.miss_deadtime++;
		log_trigger(WRTD_LOG_MISSED, WRTD_MISS_DEAD_TIME, out, &ent);
		return;
	}

	if (!wr_is_timing_ok()) {
		rule->misses++;
		out->stats.miss_no_timing++;
		log_trigger(WRTD_LOG_MISSED, WRTD_MISS_NO_WR, out, &ent);
		return;
	}

	switch (out->config.state) {
	case OUT_ST_IDLE: /* output not armed */
		return;

	case OUT_ST_ARMED:
		if (rule->state & HASH_ENT_CONDITION) {
			/* FIXME select the correct channel */
			out->priv.pending_trig = &triggers[rule->cond_ptr].ocfg[0];
			out->config.state = OUT_ST_CONDITION_HIT;
			return;
		} else if (out->config.mode == WRTD_TRIGGER_MODE_SINGLE) {
			out->config.flags &= ~WRTD_ARMED;
			out->config.state = OUT_ST_IDLE;
		}
		break;
	case OUT_ST_TEST_PENDING:
		return;

	case OUT_ST_CONDITION_HIT:
		if (rule != out->priv.pending_trig)
			return;

		if (out->config.mode == WRTD_TRIGGER_MODE_SINGLE) {
			out->config.flags &= ~WRTD_ARMED;
			out->config.state = OUT_ST_IDLE;
		} else {
			out->config.state = OUT_ST_ARMED;
		}
		break;
	default:
		break;

	}

	pq_ent = pulse_queue_push(&out->priv.queue);
	if (!pq_ent) {
		ent.ts = *ts;
		ent.id = *id;
		ent.seq = seq;
		log_trigger (WRTD_LOG_MISSED, WRTD_MISS_OVERFLOW, out, &ent);

		out->stats.miss_overflow++;
		return;
	}

	pq_ent->trig.ts = adjusted;
	pq_ent->trig.id = *id;
	pq_ent->trig.seq = seq;
	pq_ent->origin_cycles = ts->ticks;
	pq_ent->rule = rule;

	out->stats.last_enqueued.ts = *ts;
	out->stats.last_enqueued.id = *id;
	out->stats.last_enqueued.seq = seq;

	out->priv.prev_pulse = adjusted;

	log_trigger(WRTD_LOG_FILTERED, 0, out, &out->stats.last_enqueued);

}


static void filter_trigger(struct wrtd_trigger_entry *trig)
{
	struct wrtd_out_trigger *ent = rtfd_trigger_entry_find(&trig->id);
	int j;

	log_trigger(WRTD_LOG_PROMISC, 0, NULL, trig);
	wrtd_out_device.last_received = *trig;

	pr_debug("%s:%d Trigger %"PRId32":%"PRId32":%"PRId32" - entry %p\n\r",
		  __func__, __LINE__,
		  trig->id.system, trig->id.source_port, trig->id.trigger, ent);
	if (!ent)
		return;

	for(j = 0; j < FD_NUM_CHANNELS; j++)
		if(ent->ocfg[j].state != HASH_ENT_EMPTY)
			enqueue_trigger(j, &ent->ocfg[j],
					&trig->id, &trig->ts, trig->seq);

}

static void do_rx(void)
{
	volatile struct wrtd_trigger_entry *ent;
	struct wrtd_trigger_entry trig;
	struct wrtd_trigger_message *msg;
	int i;

	/* check from etherbone */
	if (mq_poll_in(TRTL_RMQ, 1 << WRTD_OUT_RMQ)) {
		msg = mq_map_in_buffer(TRTL_RMQ, WRTD_OUT_RMQ);
		for (i = 0; i < msg->count; i++)
			filter_trigger(&msg->triggers[i]);

		mq_discard(TRTL_RMQ, WRTD_OUT_RMQ);
		wrtd_out_device.counter_etherbone++;
	}

	/* Check from the loopback */
	ent = loop_queue_pop();
	if (ent) {
		trig = *ent;
		filter_trigger (&trig);
		wrtd_out_device.counter_loopback++;
	}
}

static void do_outputs(void)
{
	int i;

	for (i = 0;i < FD_NUM_CHANNELS; i++)
		do_output(&wrtd_out_channels[i]);
}

/*.
 * WRTD Command Handlers
 */

struct wrtd_trigger_sw {
	uint32_t chan;
	uint32_t now;
	struct wr_timestamp tstamp;
};


/**
 * It generate a software trigger accorging to the trigger entry coming
 * from the user space.
 */
static int wrtd_out_trigger_sw(struct trtl_fw_msg *msg_i,
			       struct trtl_fw_msg *msg_o)
{
	struct wrtd_trigger_sw *tsw = msg_i->payload;
	struct pulse_queue_entry *pq_ent;

	if (tsw->now) {
		/* Prepare a pulse at (now + some margin) */
		tsw->tstamp.seconds = lr_readl(MT_CPU_LR_REG_TAI_SEC);
		tsw->tstamp.ticks = lr_readl(MT_CPU_LR_REG_TAI_CYCLES);
		tsw->tstamp.frac = 0;
		tsw->tstamp.ticks += 10000;
	}

	/* clear all pulses from the queue */
	pulse_queue_init(&wrtd_out_channels[tsw->chan].priv.queue);
	pq_ent = pulse_queue_push (&wrtd_out_channels[tsw->chan].priv.queue);

	pq_ent->trig.ts = tsw->tstamp;
	pq_ent->trig.id.system = 0xffffffff;
	pq_ent->trig.id.source_port = 0xffffffff;
	pq_ent->trig.id.trigger = 0xffffffff;
	pq_ent->trig.seq = 0xffffffff;
	pq_ent->origin_cycles = tsw->tstamp.ticks;
	pq_ent->rule = NULL;

	wrtd_out_channels[tsw->chan].stats.last_enqueued.ts = tsw->tstamp;
	wrtd_out_channels[tsw->chan].stats.last_enqueued.id = pq_ent->trig.id;
	wrtd_out_channels[tsw->chan].stats.last_enqueued.seq = pq_ent->trig.seq;
	wrtd_out_channels[tsw->chan].config.state = OUT_ST_TEST_PENDING;

	return 0;
}


/**
 * It gets the first available row in the hash table for the given
 * trigger ID
 */
static int wrtd_out_hash_table_free(struct wrtd_trig_id *tid)
{
	int i, hidx = wrtd_hash_func(tid);

	for (i = 0; i < FD_HASH_ENTRIES;
	     ++i, hidx = (hidx + 1) & (FD_HASH_ENTRIES - 1)) {
		if (!ht[hidx])
			return hidx;
		/* Check if it already exists */
		if (trig_eq(tid, &ht[hidx]->id))
			break;
	}
	return -1;
}


/**
 * Insert trigger entry in the hash table
 */
static int wrtd_out_hash_table_insert(struct trtl_fw_msg *msg_i,
				      struct trtl_fw_msg *msg_o)
{
	uint32_t tidx = *(uint32_t *)msg_i->payload;
	int hidx;

	if (tlist_count >= FD_HASH_ENTRIES) {
		pr_error("Too many triggers %d/%d\n\r",
			  tlist_count, FD_HASH_ENTRIES);
		return -1;
	}
	if (!(triggers[tidx].flags & ENTRY_FLAG_VALID))
		return 0; /* nothing to do is a valid trigger */

	hidx = wrtd_out_hash_table_free(&triggers[tidx].id);
	if (hidx < 0 || hidx >= FD_HASH_ENTRIES)
		return -1;

	tlist_count++;
	ht[hidx] = &triggers[tidx];
	msg_o->header->msg_id = msg_i->header->msg_id;

	return 0;
}


/**
 * Remove trigger entry from the hash table
 */
static int wrtd_out_hash_table_remove(struct trtl_fw_msg *msg_i,
				      struct trtl_fw_msg *msg_o)
{
	uint32_t tidx = *(uint32_t *)msg_i->payload;
	int hidx;

	if (tlist_count <= 0) {
		pr_error("No trigger to remove %d/%d\n\r",
			  tlist_count, FD_HASH_ENTRIES);
		return -1;
	}

	if (triggers[tidx].flags & ENTRY_FLAG_VALID)
		return 0; /* nothing to do is a valid trigger */

	hidx = wrtd_out_hash_table_find(&triggers[tidx].id);
	if (hidx < 0)
		return -1;

	tlist_count--;
	ht[hidx] = NULL;

	msg_o->header->msg_id = msg_i->header->msg_id;

	return 0;
}


/**
 * It gets the trigger index for a given trigger ID
 */
static int wrtd_out_trigger_index(struct trtl_fw_msg *msg_i,
				  struct trtl_fw_msg *msg_o)
{
	struct wrtd_trig_id *id = msg_i->payload;
	uint32_t *index = msg_o->payload;
	int hidx;

	/* Verify that the size is correct */
	if (msg_i->header->len * 4 != sizeof(struct wrtd_trig_id)) {
		pr_error("%s: wrong incoming message size %d (expected %d)\n\r",
			  __func__, msg_i->header->len * 4,
			  sizeof(struct wrtd_trig_id));
		goto err;
	}


	hidx = wrtd_out_hash_table_find(id);
	if (hidx >= 0) {
		*index = ((int)ht[hidx] - (int)triggers) /
			sizeof(struct wrtd_out_trigger);
	} else {
		/* Get first free */
		for (*index = 0; *index < FD_HASH_ENTRIES; ++(*index)) {
			if (!(triggers[*index].flags & ENTRY_FLAG_VALID)) {
				msg_o->header->msg_id = WRTD_OUT_ACTION_TRIG_FRE;
				break;
			}
		}
		if ((*index) == FD_HASH_ENTRIES)
			return -1;
	}

	msg_o->header->msg_id = WRTD_OUT_ACTION_TRIG_IDX;
	msg_o->header->len = 1;
	return 0;

err:
	return -1;
}


/**
 * It disable the given channel and clear its internal queue
 */
static int wrtd_out_disable(struct trtl_fw_msg *msg_i,
			    struct trtl_fw_msg *msg_o)
{
	uint32_t ch = ((uint32_t *)msg_i->payload)[0];

	pulse_queue_init(&wrtd_out_channels[ch].priv.queue);
	fd_ch_writel(&wrtd_out_channels[ch], FD_DCR_MODE, FD_REG_DCR);

	msg_o->header->msg_id = msg_i->header->msg_id;

	return 0;
}


/**
 * Data structures to export to host system. Initialized dynamically in
 * the function init()
 */
static struct trtl_fw_buffer wrtd_out_buffers[__WRTD_OUT_STRUCT_MAX
					       + FD_HASH_ENTRIES] = {
	[OUT_STRUCT_DEVICE] = {
		.buf = &wrtd_out_device,
		.len = sizeof(struct wrtd_out),
	},
};

static struct trtl_fw_variable wrtd_out_variables[] = {
	[OUT_VAR_DEVICE_TIME_S] = {
		.addr = TRTL_ADDR_LR(MT_CPU_LR_REG_TAI_SEC),
		.mask = 0xFFFFFFFF,
		.offset = 0,
	},
	[OUT_VAR_DEVICE_TIME_T] = {
		.addr = TRTL_ADDR_LR(MT_CPU_LR_REG_TAI_CYCLES),
		.mask = 0xFFFFFFFF,
		.offset = 0,
	},
};

static trtl_fw_action_t *wrtd_out_actions[] = {
	[WRTD_OUT_ACTION_SW_TRIG] = wrtd_out_trigger_sw,
	[WRTD_OUT_ACTION_TRIG_IDX] = wrtd_out_trigger_index,
	[WRTD_OUT_ACTION_TRIG_ADD] = wrtd_out_hash_table_insert,
	[WRTD_OUT_ACTION_TRIG_DEL] = wrtd_out_hash_table_remove,
	[WRTD_OUT_ACTION_DISABLE] = wrtd_out_disable,
};


static int wrtd_o_sim_init(void)
{
	int i;

	for (i = 0; i < FD_NUM_CHANNELS; i++) {
		int hidx;
		wrtd_out_channels[i].config.log_level = WRTD_LOG_ALL;
		wrtd_out_channels[i].config.state = OUT_ST_ARMED;
		wrtd_out_channels[i].config.flags =
			WRTD_ENABLED | WRTD_ARMED | WRTD_TRIGGER_ASSIGNED;
		triggers[i].id.system = 0;
		triggers[i].id.source_port = i;
		triggers[i].id.trigger = i;
		triggers[i].flags = ENTRY_FLAG_VALID;
		triggers[i].ocfg[i].delay_cycles = 0;
		triggers[i].ocfg[i].delay_frac = 0;
		triggers[i].ocfg[i].state = HASH_ENT_DIRECT;
		hidx = wrtd_out_hash_table_free(&triggers[i].id);
		tlist_count++;
		ht[hidx] = &triggers[i];
	}

	return 0;
}

/**
 * Initialize data structures, RT application and variables
 */
static int wrtd_o_init(void)
{
	struct trtl_ep_eth_address addr = {
		.type = TRTL_EP_FRAME_UDP,
		.dst_mac = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF},
		.dst_port = WRTD_OUT_PORT,
		.ethertype = 0x800, /* IPv4 */
		.filter = TRTL_EP_FILTER_UDP |
			  TRTL_EP_FILTER_DST_PORT,
	};
	int i, j;

	addr.filter |= TRTL_EP_FILTER_ENABLE;

	if (!wr_present()) {
		pr_error("WhiteRabbit not found\n\r");
		return -EINVAL;
	}

	tlist_count = 0;
	wr_state = WR_LINK_OFFLINE;
	wr_enable_lock(0);

	/* device */
	memset(&wrtd_out_device, 0, sizeof(struct wrtd_out));

	/* Channels */
	for (i = 0; i < FD_NUM_CHANNELS; i++) {
		memset(&wrtd_out_channels[i], 0,
		       sizeof(struct wrtd_out_channel));
		wrtd_out_channels[i].base_addr = 0x100 + i * 0x100;
		wrtd_out_channels[i].n = i;
		wrtd_out_channels[i].config.mode = WRTD_TRIGGER_MODE_AUTO;
		wrtd_out_channels[i].priv.pending_trig = NULL;
		wrtd_out_channels[i].config.state = OUT_ST_IDLE;
		wrtd_out_channels[i].priv.idle = 1;
		wrtd_out_channels[i].config.dead_time = 80000 / 8; // 80 us
		wrtd_out_channels[i].config.width_cycles = 1250; // 1us
		pr_debug("mode %d %"PRId32"\n\r", i , wrtd_out_channels[i].config.mode);
	}

	/* Triggers */
	for (i = 0; i < FD_HASH_ENTRIES; i++) {
		memset(&triggers[i], 0, sizeof(struct wrtd_out_trigger));
		ht[i] = NULL;
	}

	/* librt: dynamically add structures */
	for (i = 0, j = 1; i < FD_NUM_CHANNELS; i++, j++) {
		wrtd_out_buffers[j].buf = &wrtd_out_channels[i];
		wrtd_out_buffers[j].len = sizeof(struct wrtd_out_channel)
			- sizeof(struct wrtd_out_channel_private);
	}
	for (i = 0, j = __WRTD_OUT_STRUCT_MAX; i < FD_HASH_ENTRIES; i++, j++) {
		wrtd_out_buffers[j].buf = &triggers[i];
		wrtd_out_buffers[j].len = sizeof(struct wrtd_out_trigger);
	}

	rmq_bind_in(WRTD_OUT_RMQ, TRTL_EP_ETH, &addr);

	/* Skip WR sync and automatically generate some events when simulating */
	if (is_simulation)
		wrtd_o_sim_init();

	if (!is_simulation)
		fd_init();

	pr_debug("rt-output firmware initialized.\n\r");

	return 0;
}

static int wrtd_o_main(void)
{
	while (1) {
		do_rx();
		do_outputs();
		trtl_fw_mq_action_dispatch(TRTL_HMQ, WRTD_OUT_HMQ);
		wr_update_link();
	}

	return 0;
}

struct trtl_fw_application app = {
	.name = "wrtd-output",
	.version = {
		.rt_id = WRTD_OUT_RT_ID,
		.rt_version = RT_VERSION(2, 0),
		.git_version = GIT_VERSION
	},

	.buffers = wrtd_out_buffers,
	.n_buffers = ARRAY_SIZE(wrtd_out_buffers),

	.variables = wrtd_out_variables,
	.n_variables = ARRAY_SIZE(wrtd_out_variables),

	.actions = wrtd_out_actions,
	.n_actions = ARRAY_SIZE(wrtd_out_actions),

	.init = wrtd_o_init,
	.main = wrtd_o_main,
	.exit = NULL,
};
